package main

import (
	"context"
	"flag"
	"fmt"
	"log"
	"net/http"
	"net/url"
	"os"
	"os/signal"
	"path"
	"path/filepath"
	"strings"
	"syscall"
	"text/template"

	"srs.epita.fr/fic-server/admin/api"
	"srs.epita.fr/fic-server/admin/pki"
	"srs.epita.fr/fic-server/admin/sync"
	"srs.epita.fr/fic-server/libfic"
	"srs.epita.fr/fic-server/settings"
)

var StaticDir string

type ResponseWriterPrefix struct {
	real   http.ResponseWriter
	prefix string
}

func (r ResponseWriterPrefix) Header() http.Header {
	return r.real.Header()
}

func (r ResponseWriterPrefix) WriteHeader(s int) {
	if v, exists := r.real.Header()["Location"]; exists {
		r.real.Header().Set("Location", r.prefix+v[0])
	}
	r.real.WriteHeader(s)
}

func (r ResponseWriterPrefix) Write(z []byte) (int, error) {
	return r.real.Write(z)
}

func StripPrefix(prefix string, h http.Handler) http.Handler {
	if prefix == "" {
		return h
	}
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if prefix != "/" && r.URL.Path == "/" {
			http.Redirect(w, r, prefix+"/", http.StatusFound)
		} else if p := strings.TrimPrefix(r.URL.Path, prefix); len(p) < len(r.URL.Path) {
			r2 := new(http.Request)
			*r2 = *r
			r2.URL = new(url.URL)
			*r2.URL = *r.URL
			r2.URL.Path = p
			h.ServeHTTP(ResponseWriterPrefix{w, prefix}, r2)
		} else {
			h.ServeHTTP(w, r)
		}
	})
}

func main() {
	cloudDAVBase := ""
	cloudUsername := "fic"
	cloudPassword := ""
	localImporterDirectory := ""
	localImporterSymlink := false

	// Read paremeters from environment
	if v, exists := os.LookupEnv("FICCA_PASS"); exists {
		pki.SetCAPassword(v)
	} else {
		log.Println("WARNING: no password defined for the CA, will use empty password to secure CA private key")
		log.Println("WARNING:    PLEASE DEFINE ENVIRONMENT VARIABLE: FICCA_PASS")
	}
	if v, exists := os.LookupEnv("FICCLOUD_URL"); exists {
		cloudDAVBase = v
	}
	if v, exists := os.LookupEnv("FICCLOUD_USER"); exists {
		cloudUsername = v
	}
	if v, exists := os.LookupEnv("FICCLOUD_PASS"); exists {
		cloudPassword = v
	}

	// Read parameters from command line
	var bind = flag.String("bind", "127.0.0.1:8081", "Bind port/socket")
	var dsn = flag.String("dsn", fic.DSNGenerator(), "DSN to connect to the MySQL server")
	var baseURL = flag.String("baseurl", "/", "URL prepended to each URL")
	flag.StringVar(&api.TimestampCheck, "timestampCheck", api.TimestampCheck, "Path regularly touched by frontend to check time synchronisation")
	flag.StringVar(&pki.PKIDir, "pki", "./PKI", "Base directory where found PKI scripts")
	flag.StringVar(&StaticDir, "static", "./htdocs-admin/", "Directory containing static files")
	flag.StringVar(&api.TeamsDir, "teams", "./TEAMS", "Base directory where save teams JSON files")
	flag.StringVar(&api.DashboardDir, "dashbord", "./DASHBOARD", "Base directory where save public JSON files")
	flag.StringVar(&settings.SettingsDir, "settings", settings.SettingsDir, "Base directory where load and save settings")
	flag.StringVar(&fic.FilesDir, "files", fic.FilesDir, "Base directory where found challenges files, local part")
	flag.StringVar(&localImporterDirectory, "localimport", localImporterDirectory,
		"Base directory where found challenges files to import, local part")
	flag.BoolVar(&localImporterSymlink, "localimportsymlink", localImporterSymlink,
		"Copy files or just create symlink?")
	flag.StringVar(&cloudDAVBase, "clouddav", cloudDAVBase,
		"Base directory where found challenges files to import, cloud part")
	flag.StringVar(&cloudUsername, "clouduser", cloudUsername, "Username used to sync")
	flag.StringVar(&cloudPassword, "cloudpass", cloudPassword, "Password used to sync")
	flag.BoolVar(&fic.OptionalDigest, "optionaldigest", fic.OptionalDigest, "Is the digest required when importing files?")
	flag.BoolVar(&fic.StrongDigest, "strongdigest", fic.StrongDigest, "Are BLAKE2b digests required or is SHA-1 good enough?")
	flag.Parse()

	log.SetPrefix("[admin] ")

	// Instantiate importer
	if localImporterDirectory != "" && cloudDAVBase != "" {
		log.Fatal("Cannot have both --clouddav and --localimport defined.")
		return
	} else if localImporterDirectory != "" {
		sync.GlobalImporter = sync.LocalImporter{Base: localImporterDirectory, Symlink: localImporterSymlink}
	} else if cloudDAVBase != "" {
		sync.GlobalImporter, _ = sync.NewCloudImporter(cloudDAVBase, cloudUsername, cloudPassword)
	}
	if sync.GlobalImporter != nil {
		log.Println("Using", sync.GlobalImporter.Kind())
	}

	// Sanitize options
	var err error
	log.Println("Checking paths...")
	if StaticDir, err = filepath.Abs(StaticDir); err != nil {
		log.Fatal(err)
	}
	sync.DeepReportPath = path.Join(StaticDir, sync.DeepReportPath)
	if fic.FilesDir, err = filepath.Abs(fic.FilesDir); err != nil {
		log.Fatal(err)
	}
	if pki.PKIDir, err = filepath.Abs(pki.PKIDir); err != nil {
		log.Fatal(err)
	}
	if api.DashboardDir, err = filepath.Abs(api.DashboardDir); err != nil {
		log.Fatal(err)
	}
	if api.TeamsDir, err = filepath.Abs(api.TeamsDir); err != nil {
		log.Fatal(err)
	}
	if api.TimestampCheck, err = filepath.Abs(api.TimestampCheck); err != nil {
		log.Fatal(err)
	}
	if settings.SettingsDir, err = filepath.Abs(settings.SettingsDir); err != nil {
		log.Fatal(err)
	}
	if *baseURL != "/" {
		tmp := path.Clean(*baseURL)
		baseURL = &tmp
	} else {
		tmp := ""
		baseURL = &tmp
	}

	// Creating minimal directories structure
	os.MkdirAll(fic.FilesDir, 0777)
	os.MkdirAll(pki.PKIDir, 0711)
	os.MkdirAll(api.TeamsDir, 0777)
	os.MkdirAll(api.DashboardDir, 0777)
	os.MkdirAll(settings.SettingsDir, 0777)

	// Initialize settings and load them
	if !settings.ExistsSettings(path.Join(settings.SettingsDir, settings.SettingsFile)) {
		if err = settings.SaveSettings(path.Join(settings.SettingsDir, settings.SettingsFile), settings.FICSettings{
			Title:                  "Challenge FIC",
			Authors:                "Laboratoire SRS, ÉPITA",
			FirstBlood:             fic.FirstBlood,
			SubmissionCostBase:     fic.SubmissionCostBase,
			ExerciceCurCoefficient: 1,
			HintCurCoefficient:     1,
			WChoiceCurCoefficient:  1,
			AllowRegistration:      false,
			CanJoinTeam:            false,
			DenyNameChange:         false,
			AcceptNewIssue:         true,
			EnableResolutionRoute:  false,
			PartialValidation:      true,
			UnlockedChallengeDepth: 0,
			SubmissionUniqueness:   false,
			DisplayAllFlags:        false,
			EventKindness:          false,
		}); err != nil {
			log.Fatal("Unable to initialize settings.json:", err)
		}
	} else {
		var config settings.FICSettings
		if config, err = settings.ReadSettings(path.Join(settings.SettingsDir, settings.SettingsFile)); err != nil {
			log.Fatal("Unable to read settings.json:", err)
		} else {
			api.ApplySettings(config)
		}
	}

	// Database connection
	log.Println("Opening database...")
	if err = fic.DBInit(*dsn); err != nil {
		log.Fatal("Cannot open the database: ", err)
	}
	defer fic.DBClose()

	log.Println("Creating database...")
	if err = fic.DBCreate(); err != nil {
		log.Fatal("Cannot create database: ", err)
	}

	// Update base URL on main page
	log.Println("Changing base URL to", *baseURL+"/", "...")
	if file, err := os.OpenFile(path.Join(StaticDir, "index.html"), os.O_CREATE|os.O_WRONLY|os.O_TRUNC, os.FileMode(0644)); err != nil {
		log.Println("Unable to open index.html: ", err)
	} else if indexTmpl, err := template.New("index").Parse(indextpl); err != nil {
		log.Println("Cannot create template: ", err)
	} else if err = indexTmpl.Execute(file, map[string]string{"urlbase": path.Clean(path.Join(*baseURL+"/", "nuke"))[:len(path.Clean(path.Join(*baseURL+"/", "nuke")))-4]}); err != nil {
		log.Println("An error occurs during template execution: ", err)
	}

	// Prepare graceful shutdown
	interrupt := make(chan os.Signal, 1)
	signal.Notify(interrupt, os.Interrupt, syscall.SIGTERM)

	srv := &http.Server{
		Addr:    *bind,
		Handler: StripPrefix(*baseURL, api.Router()),
	}

	// Serve content
	go func() {
		log.Fatal(srv.ListenAndServe())
	}()
	log.Println(fmt.Sprintf("Ready, listening on %s", *bind))

	// Wait shutdown signal
	<-interrupt

	log.Print("The service is shutting down...")
	srv.Shutdown(context.Background())
	log.Println("done")
}
